#include "Arduino.h"
#include <SPI.h>
#include "TLC5615.h"

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#define BUFFER_SIZE 120
#define FLASH_SIZE 8 //*1Mbit
#define FLASH_HOLD 8
#define FLASH_CS 9
#define DAC_CS  10

uint8_t command[4] = {0x00, 0x00, 0x00, 0x00};
bool flash_ava = true, new_data = false, dump = false;
unsigned long timer;
uint16_t address_counter = 0, page_count = 0, page_size = 4096;
TLC5615 dac(10);
uint16_t s_buffer[BUFFER_SIZE],head = 0, pos = 0;
uint16_t analogData;
char option;

void Timer2INIT()
{
    noInterrupts();
    TCCR2A = 0;
    TCCR2B = 0;
    TCNT1 = 0;

    OCR2A = 100;
    TCCR2A |= (1 << WGM21);
    TCCR2B |= (1 << CS21);
    TIMSK2 |= (1 << OCIE2A);
    interrupts();
}

ISR(TIMER2_COMPA_vect)
{
    analogData = analogRead(0);
    new_data = true; //A flag whether the new analog data is available to read.
}

void setup() {
  // put your setup code here, to run once:
    pinMode(DAC_CS, OUTPUT);
    pinMode(FLASH_HOLD, OUTPUT);
    pinMode(FLASH_CS, OUTPUT);
    digitalWrite(DAC_CS, HIGH);
    digitalWrite(FLASH_CS, HIGH);
    digitalWrite(FLASH_HOLD, HIGH);
    Serial.begin(115200);
    SPI.setClockDivider(SPI_CLOCK_DIV8);
    SPI.begin();
    dac.begin();
    sbi(ADCSRA,ADPS2) ;
    cbi(ADCSRA,ADPS1) ;
    sbi(ADCSRA,ADPS0) ;
    Timer2INIT();
    
  //Serial.println("start");
  
  //FlashStartWrite(page_count);
  
  
}

void loop() {
    Serial.println("CENG4480 LAB5");
    Serial.println("1. Write");
    Serial.println("2. Read");
    Serial.println("3. Clear All Data");
    Serial.println("Select your option: ");
    while(!Serial.available());
    option = Serial.read();
    String address;
    int page;
    switch(option)
    {
      case '1':{
        Serial.println("Enter the page number you want to write: ");
        while(!Serial.available());
        address = Serial.readString();
        Serial.println("Enter the Data to want to write (within 256 bytes long): ");
        /*
        Please input your code here.
        */
        while(CheckForFinish() == 0){}
        Serial.println("DONE");
        break;
      }
      case '2':{
        Serial.println("Enter the page number to want to read: ");
        while(!Serial.available());
        address = Serial.readString();
        Serial.println("Enter the length of data you want to read: ");
        /*
        Please input your code here.
        */
          break;
          }
      case '3':{
        /*
        Please input your code here.
        */
        Serial.println("DONE");
        break;
     }
    }
    /* This part is only for Procedure 6.
    digitalWrite(DAC_CS, LOW);//enable the DAC chip
    if(new_data) {
      dac.analogWrite(analogData); // wirte the data to DAC chip
      new_data = false;
    }    
   */
}

//-------------------API area------------------//

void FlashStartWrite(uint32_t page)
{
    uint32_t address;
    uint8_t result;
    EnableWrite();
    WaitForEnable();
    digitalWrite(FLASH_CS, LOW);
    address = page * 256;
    command[0] = 0x02;
    command[1] = (address >> 16);
    command[2] = (address >> 8);
    command[3] = address_counter;
  SPI.transfer(command,4);
    delayMicroseconds(10);
}

void FlashStoreByte(uint8_t input)
{
  SPI.transfer(input);
}

void FlashStartRead(uint32_t page)
{
    uint32_t address;
    uint8_t result;
    address = page * 256;
    digitalWrite(FLASH_CS, LOW);
    command[0] = 0x03;
    command[1] = (address >> 16) & 0xFF;
    command[2] = (address >> 8) & 0xFF;
    command[3] = address & 0xFF;
    SPI.transfer(command,4);
    delayMicroseconds(10);
}

uint8_t FlashReadByte()
{
  uint8_t result;
  result = SPI.transfer(0x00);
  //Serial.println(result);
 
  return result;
}

void FlashEndAction()
{
    digitalWrite(FLASH_CS, HIGH);
}

void DeleteAll()
{
  EnableWrite();
  digitalWrite(FLASH_CS, LOW);
  SPI.transfer(0x60);
  digitalWrite(FLASH_CS, HIGH);
  WaitForFinish();
}

void WaitForFinish()
{
  uint8_t result;
  digitalWrite(FLASH_CS, LOW);
  result = SPI.transfer(0x05);
  SPI.transfer(0x00);
  while(result & 0x01 == 1)
  {
    result = SPI.transfer(0x05);
  }
  digitalWrite(FLASH_CS, HIGH);
}

boolean CheckForFinish()
{
  uint8_t result;
  digitalWrite(FLASH_CS, LOW);
  SPI.transfer(0x05);
  result = SPI.transfer(0x00);
  digitalWrite(FLASH_CS, HIGH);
  if(result & 0x01 == 1)
    return false;
  else
    return true;
  
}

void EnableWrite()
{
  digitalWrite(FLASH_CS, LOW);
  SPI.transfer(0x06);
  digitalWrite(FLASH_CS, HIGH);
}

//-------------unused API functions--------------//

void addBuffer(uint16_t temp)
{
  s_buffer[pos] = temp;
  pos++;
  if(pos == BUFFER_SIZE)
    pos = 0;
}

bool popBuffer(uint16_t* output)
{
  if(head == pos)
  {
    //SPI.transfer16(head);
    head = pos = 0;
    
    return false;
  }
  *output = s_buffer[head];
  head++;
  if(head == BUFFER_SIZE)
    head = 0;
  return true;
}

void FlashStoreWord(uint16_t input)
{
  SPI.transfer16(input);
}

uint16_t FlashReadWord()
{
  uint8_t result;
  uint16_t flash_read = 0;
  result = SPI.transfer(0x00);
  //Serial.println(result);
  flash_read |= result << 8;
  result = SPI.transfer(0x00);
  //Serial.println(result);
  flash_read |= result;
  return flash_read;
}

void ClearFlash()
{
    EnableWrite();
    WaitForEnable();
    command[0] = 0x20;
    digitalWrite(FLASH_CS, LOW);
    SPI.transfer(command,4);
    digitalWrite(FLASH_CS, HIGH);
    WaitForFinish();
}

void WaitForEnable()
{
    uint8_t result;
    digitalWrite(FLASH_CS, LOW);
    SPI.transfer(0x05);
    result = SPI.transfer(0x00);
    while(result & 0x10 == 1)
    {
        SPI.transfer(0x05);
        result = SPI.transfer(0x00);
    }
    digitalWrite(FLASH_CS, HIGH);
}
